#!/usr/bin/python

import os
import shutil
import sys
import base64
import subprocess
import yaml
from charmhelpers.contrib.charmsupport.nrpe import NRPE
try:
    from jinja2 import Template
except ImportError:
    pass

from charmhelpers.core.hookenv import (
    Hooks,
    UnregisteredHookError,
    config,
    log,
    open_port,
)

from charmhelpers.core import (
    sysctl,
)

from charmhelpers.fetch import (
    apt_install,
    apt_update,
)

from charmhelpers.contrib.openstack.utils import (
    configure_installation_source,
)

from charmhelpers.payload.execd import execd_preinstall

default_nagios_plugin_dir = "/usr/lib/nagios/plugins"


hooks = Hooks()

BASE_PACKAGES = [
    'openvpn',
    'python-jinja2',
    'sudo',
]


# FIXME
def determine_packages():
    apt_extra_packages = config('apt_extra_packages')
    if apt_extra_packages:
        packages = [] + BASE_PACKAGES + apt_extra_packages
    else:
        packages = [] + BASE_PACKAGES
    return list(set(packages))


def install_nrpe_scripts():
    if not os.path.exists(default_nagios_plugin_dir):
        os.makedirs(default_nagios_plugin_dir)
    shutil.copy2('{}/files/check_openvpn'.format(os.environ['CHARM_DIR']),
                 '{}/check_openvpn'.format(default_nagios_plugin_dir))


def update_nrpe_config():
    old_mask = os.umask(02)
    conf = config()
    nrpe_compat = NRPE()
    check = False
    for daemon in ['udp', 'tcp']:
        if conf['{}_daemon'.format(daemon)]:
            vpn_name = conf['{}_proto'.format(daemon)]
            nrpe_compat.add_check(
                shortname="openvpn-{}".format(vpn_name),
                description="Check {} openvpn".format(vpn_name),
                check_cmd="check_openvpn {}".format(vpn_name)
            )
            check = True
    if check:
        nrpe_compat.write()
    os.umask(old_mask)


@hooks.hook('config-changed')
def config_changed():
    conf = config()

    os.umask(07)
    if subprocess.call(['getent', 'group', conf['group']]) != 0:
        subprocess.call(['addgroup', conf['group']])
    if subprocess.call(['getent', 'passwd', conf['user']]) != 0:
        subprocess.call(['adduser', '--gecos', 'openvpn,,,,', '--disabled-login', '--ingroup', conf['group'], conf['user']])

    open("/etc/openvpn/ca.crt", "w").write(str(base64.b64decode(conf['ca_certificate'])))
    open("/etc/openvpn/server.key", "w").write(str(base64.b64decode(conf['server_key'])))
    open("/etc/openvpn/server.crt", "w").write(str(base64.b64decode(conf['server_certificate'])))
    open("/etc/openvpn/ta.key", "w").write(str(base64.b64decode(conf['tls_auth_key'])))
    if not os.access('/etc/openvpn/dh.pem', os.R_OK):
        subprocess.call(['openssl', 'gendh', '-out', '/etc/openvpn/dh.pem'])

    subprocess.call(['chown', '{}:'.format(conf['user']), '/etc/openvpn/ta.key'])

    if conf['udp_daemon']:
        if conf['udp_ip_address'] == '':
            if conf['udp_proto'] == 'udp':
                cmd = "ip route get 8.8.8.8 | sed -n '/8.8.8.8/s/.*src //p'"
            else:
                cmd = "ip route get 2001::1 | sed -n '/^2/s/.*src \([0-9a-f:]*\) .*$/\\1/p'"
            conf['udp_ip_address'] = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE).stdout.read().strip()
        conf['conf_cmds'] = yaml.safe_load(conf['udp_extra_commands'])
        udp_conf = Template(open("templates/udp.conf.tmpl").read()).render(conf)
        with open("/etc/openvpn/{}.conf".format(conf['udp_proto']), "w") as udp_file:
            udp_file.write(str(udp_conf))
        if conf['udp_client_config_directory'] != '':
            try:
                print "creating udp ccd: {}".format(conf['udp_client_config_directory'])
                os.makedirs(conf['udp_client_config_directory'])
            except OSError:
                pass
            subprocess.call(['chown', '-R', '{}:'.format(conf['user']), conf['udp_client_config_directory']])
        open_port(conf['udp_port'], 'udp')
    else:
        try:
            os.unlink('/etc/openvpn/{}.conf'.format(conf['udp_proto']))
        except OSError:
            pass

    if conf['tcp_daemon']:
        if conf['tcp_ip_address'] == '':
            if conf['tcp_proto'] == 'tcp':
                cmd = "ip route get 8.8.8.8 | sed -n '/8.8.8.8/s/.*src //p'"
            else:
                cmd = "ip route get 2001::1 | sed -n '/^2/s/.*src \([0-9a-f:]*\) .*$/\\1/p'"
            conf['tcp_ip_address'] = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE).stdout.read().strip()
        conf['conf_cmds'] = yaml.safe_load(conf['tcp_extra_commands'])
        tcp_conf = Template(open("templates/tcp.conf.tmpl").read()).render(conf)
        with open("/etc/openvpn/{}.conf".format(conf['tcp_proto']), "w") as tcp_file:
            tcp_file.write(str(tcp_conf))
        if conf['tcp_client_config_directory'] != '':
            try:
                print "creating tcp ccd: {}".format(conf['tcp_client_config_directory'])
                os.makedirs(conf['tcp_client_config_directory'])
            except OSError:
                pass
            subprocess.call(['chown', '-R', '{}:'.format(conf['user']), conf['tcp_client_config_directory']])
        open_port(conf['tcp_port'], 'tcp')
    else:
        try:
            os.unlink('/etc/openvpn/{}.conf'.format(conf['udp_proto']))
        except OSError:
            pass

    if conf['ccd_sync_key']:
        ssh_auth_host = Template(open("templates/auth_keys.tmpl").read()).render(conf)
        auth_dir = "/".join(conf['ssh_auth_key_path'].split('/')[:-1])
        if not os.access(auth_dir, os.X_OK):
            try:
                tmp = os.umask(022)
                os.makedirs(auth_dir)
                os.umask(tmp)
            except OSError:
                pass
        with open(conf['ssh_auth_key_path'], "w") as auth_file:
            auth_file.write(str(ssh_auth_host))
        subprocess.call(['chown', '-R', '{}:'.format(conf['user']), conf['ssh_auth_key_path']])

    if conf['crl_filename']:
        try:
            open(conf['crl_filename'])
        except IOError:
            crl_dir = os.path.dirname(conf['crl_filename'])
            if not os.path.exists(crl_dir):
                os.makedirs(crl_dir)
            subprocess.call(['touch', conf['crl_filename']])
        subprocess.call(['chown', '{}:'.format(conf['user']), conf['crl_filename']])

    sudoers = Template(open("templates/sudoers.tmpl").read()).render(conf)
    with open("/etc/sudoers.d/{}".format(conf['user']), "w") as sudoers_file:
        sudoers_file.write(str(sudoers))

    # someday, maybe consider making these tunable.
    # on precise, there is also: net.ipv4.netfilter.ip_conntrack_tcp_timeout_established: 14400
    # we need to force nf_conntrack to load, so that sysctl.create does not die.
    try:
        tmp = os.umask(022)
        os.makedirs("/etc/modules-load.d")
        os.umask(tmp)
    except OSError:
        pass
    open("/etc/modules-load.d/nf_conntrack", "w").write("nf_conntrack\n")
    subprocess.call(['modprobe', 'nf_conntrack'])
    sysctl.create('{ net.nf_conntrack_max: 2097152}', '/etc/sysctl.d/50-conntrack.conf')

    update_nrpe_config()

    # TODO
    #   determine when we need to restart the server, and in what manner
    #     minimize that


@hooks.hook('nrpe-external-master-relation-changed', 'local-monitors-relation-changed')
def nrpe_external_master_relation_changed():
    update_nrpe_config()


@hooks.hook('install', 'upgrade-charm')
def install():
    execd_preinstall()
    configure_installation_source(config('openvpn_origin'))
    apt_update()
    apt_install(determine_packages(), fatal=True)
    install_nrpe_scripts()
    shutil.copy2('{}/files/learn_address.py'.format(os.environ['CHARM_DIR']),
                 '/usr/local/sbin/learn-address')


def main():
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        log('Unknown hook {} - skipping.'.format(e))


if __name__ == '__main__':
    main()
